/* 
 * kill touchscreen
 * a tool for disabling the treo touchscreen
 * copyright (c) 2005, daniel grobe sachs <dgsachs@nekito.net>
 *
 * This is free software. See included LICENSE file for details.
 */

#include <PalmOS.h>
#include <palmOne_68K.h>
#include <68K/System/HsExt.h>
#include <Common/System/HsExtCommon.h>

#include "resource.h"

#define FTR_MODE 0
#define FTR_SIDETOGG 1
#define FTR_ROCKERTIME 2
#define FTR_SIDEHELD 3
#define FTR_PRESSTIME 4
#define FTR_STATE 5

#define MENUDELAY 10 // enable touchscreen for 200ms after rocker
#define SIDEDELAY 50 // delay for side button (500ms)

#define DEFAULTMODE  ENABLED

#define sysNotifyKeyReplacePriority -96

////////////////////////////////////////////////////////////////////////
// Globals

// Pointer to the currently-active form
static FormPtr gpForm;
UInt32 CurrentMode = ENABLED;


//// StartApplication //////////////////////////////////////////////////
// Called when application is starting

static Err StartApplication() 
{
	FrmGotoForm(MAIN);
	return 0;
}

//// StopApplication ///////////////////////////////////////////////////
// Called when application is being shut down.

static void StopApplication() 
{
	FrmCloseAllForms();
}

static void RegisterForNotifications()
{
	UInt16 cardNo;
	LocalID dbID;

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	FtrSet(CREATOR,FTR_ROCKERTIME,0);
	FtrSet(CREATOR,FTR_SIDEHELD,0);

	SysNotifyRegister(cardNo, dbID, sysNotifyEventDequeuedEvent, NULL,
			sysNotifyKeyReplacePriority, 0);
}

static void DeregisterNotifications()
{
	UInt16 cardNo;
	LocalID dbID;

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	SysNotifyUnregister(cardNo, dbID, sysNotifyEventDequeuedEvent,
			sysNotifyKeyReplacePriority);
}

static void ToggleTouchscreen(Boolean beep)
{
	UInt32 mode;

	// Toggle
	FtrGet(CREATOR,FTR_SIDETOGG,&mode);

	mode = !mode;
	FtrSet(CREATOR,FTR_SIDETOGG,mode);
	
	PrefSetAppPreferences(CREATOR,FTR_SIDETOGG,1,&mode,sizeof(mode),false);

	// Acknoweldge change
	if( beep )
		SndPlaySystemSound(sndInfo);
}

// Endian swapper
#define EndianSwap16(n) (((((UInt16) n) << 8) & 0xFF00) | \
		         ((((UInt16) n) >> 8) & 0x00FF))

// Notification handler
static void HandleSide(EventType *eventP)
{
	UInt32 mode;
	UInt32 time;
	UInt32 state;
	struct EventType newevent;

	FtrGet(CREATOR,FTR_MODE,&mode);

	if( mode == WITHOPT )
	{
		// keyDown posing as a keyUp?
		if( eventP->data.keyDown.modifiers & EndianSwap16(0x100) )
		{
			eventP->eType = EndianSwap16(nilEvent);
			return;
		}

		if( eventP->data.keyDown.modifiers & EndianSwap16(0x010) )
		{
			// Clear state
			FtrSet(CREATOR,FTR_PRESSTIME,TimGetTicks());
			FtrSet(CREATOR,FTR_STATE,0);

			// Toggle
			ToggleTouchscreen(true);

			// Eat keypress
			eventP->eType = EndianSwap16(nilEvent);
		}
		
		
		return;
	}

	if( mode != WITHSIDE ) 
		return;

	switch( eventP->data.keyDown.modifiers )
	{
		case EndianSwap16(0x148):	// Normal release, I hope
			//SndPlaySystemSound(sndError);
			FtrGet(CREATOR,FTR_STATE,&state);

			switch( state ) 
			{	
				case 0: // do nothing
					break;

				case 1: // Need to re-send the side key keyDown and keyUp

					// Requeue the keyUp event
					MemSet(&newevent,sizeof(EventType),0);
					newevent.eType = keyUpEvent;
					newevent.data.keyUp.chr = EndianSwap16(eventP->data.keyUp.chr);
					newevent.data.keyUp.keyCode = EndianSwap16(eventP->data.keyUp.keyCode);
					newevent.data.keyUp.modifiers = 0x008; // Normal release (I hope)
					EvtAddEventToQueue(&newevent);

					// Change this event to a keyDown (it won't get reprocessed)
					eventP->eType = EndianSwap16(keyDownEvent);
					eventP->data.keyDown.modifiers = EndianSwap16(0x808);
					break;

				case 2:
				case 3:
					// we did other stuff, so eat the keyUp
					eventP->eType = EndianSwap16(nilEvent);
	
					// reset state machine
					FtrSet(CREATOR,FTR_PRESSTIME,0);
					FtrSet(CREATOR,FTR_STATE,0);
					break;

			}
			break;
				
		case EndianSwap16(0x808): // standard unmodified press
			// Save the key-press time
			FtrSet(CREATOR,FTR_PRESSTIME,TimGetTicks());
			FtrSet(CREATOR,FTR_STATE,1);
			eventP->eType = EndianSwap16(nilEvent);

			break;
				
		case EndianSwap16(0x848): // held
		case EndianSwap16(0x8C8):
			FtrGet(CREATOR,FTR_PRESSTIME,&time);
			FtrGet(CREATOR,FTR_STATE,&state);
			eventP->eType = EndianSwap16(nilEvent);

			switch( state ) 
			{
				case 1:
					// 0.5 sec to toggle touchscreen state
					if( TimGetTicks() > (time + SysTicksPerSecond()/2) )
					{
						ToggleTouchscreen(true);
						FtrSet(CREATOR,FTR_STATE,2);
					}
					break;

				case 2:
					// 1 sec to launch
					if( TimGetTicks() > (time + SysTicksPerSecond()) )
					{
						UInt32 appid;
						UInt16 cardid;
						UInt32 dbid;
						UInt32 len = sizeof(appid);
						DmSearchStateType sstate;

						// set the state
						FtrSet(CREATOR,FTR_STATE,3);
								
						// put the touchscreen state back
						ToggleTouchscreen(false);
							
						// launch the requested application
						HsPrefGet(hsPrefSideHardCharAppCreator, &appid, &len);
						DmGetNextDatabaseByTypeCreator(true,&sstate,'appl',appid,true,&cardid,&dbid);
						if( dbid )
							SysUIAppSwitch(cardid, dbid, sysAppLaunchCmdNormalLaunch, NULL);
					}
					break;

				default:
					break;	
					
			}
			break;

		default:	// if we get something unrecognized, stop messing with crap
			// reset state machine
			FtrSet(CREATOR,FTR_PRESSTIME,0);
			FtrSet(CREATOR,FTR_STATE,0);

			break;

	}
}

//// PilotMain /////////////////////////////////////////////////////////
// It all starts here...

static void HandleNotification(SysNotifyParamType *np)
{
	Err err;
	UInt32 mode;
	EventType *eventP = (EventType*)np->notifyDetailsP;

	// Is this a queue event?
	if( np->notifyType != sysNotifyEventDequeuedEvent )
		return;

	// Yes, fiddle!

	switch( EndianSwap16(eventP->eType) )
	{
		case penDownEvent: 
		case penMoveEvent:
		case penUpEvent:
			err = FtrGet(CREATOR,FTR_MODE,&mode);
			if( err ) 
				break;

			// Check side button
			if( (mode == WITHSIDE) || (mode == WITHOPT) )
			{
				FtrGet(CREATOR,FTR_SIDETOGG,&mode);
				if( mode )
					break;
			}

			// Check rocker for menu fix
			FtrGet(CREATOR,FTR_ROCKERTIME,&mode);
			if( mode > TimGetTicks() )
				break;

			eventP->eType = EndianSwap16(nilEvent);
			break;

		case keyUpEvent:
		case keyDownEvent:
			switch( EndianSwap16(eventP->data.keyDown.chr) )
			{
				case 0x0134: // Rocker left
				case 0x0135: // Rocker right
				case 0x0136: // Rocker center
					FtrSet(CREATOR,FTR_ROCKERTIME,TimGetTicks()+MENUDELAY);
					break;

				case 0x161F: // Side
					HandleSide(eventP);
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
}

void touchScreenMode(int Mode)
{
	if( Mode == ENABLED )
		DeregisterNotifications();
	else
		RegisterForNotifications();

	FtrSet(CREATOR,0,Mode);
}

// Form handler
Boolean MainFormEventHandler(EventPtr event)
{
	Boolean handled = false;

	switch(event->eType)
	{
		case frmOpenEvent: 
			gpForm = FrmGetActiveForm();
			FrmSetControlGroupSelection(gpForm, 1, CurrentMode);
			FrmDrawForm(gpForm);
			handled = true;
			break;

		case frmCloseEvent:
			FrmEraseForm(gpForm);
			FrmDeleteForm(gpForm);
			gpForm = 0;
			break;

		case ctlSelectEvent:
			if( event->data.ctlSelect.on == true ) 
			{
				CurrentMode = event->data.ctlSelect.controlID;
				touchScreenMode(CurrentMode);
				PrefSetAppPreferences(CREATOR,FTR_MODE,1,&CurrentMode,
					sizeof(CurrentMode), true);
			}
			break;

		default:
			break;
	}

	return handled;
}


//// ApplicationEventHandler /////////////////////////////////////////////
// Handle global events, and pass form-specific events to each form's
// event handler.

static Boolean ApplicationEventHandler(EventPtr event) 
{
	Boolean handled = false;
	FormPtr pForm;

	switch (event->eType) {
		case frmLoadEvent: 
			pForm = FrmInitForm(event->data.frmLoad.formID);
			FrmSetActiveForm(pForm);
			FrmSetEventHandler(pForm, MainFormEventHandler);
			handled = true;
			break;

		default:
			// -Wall warning eater: switch must handle all enum
			// elements.
			break;
	}

	return handled;
}


//// EventLoop /////////////////////////////////////////////////////////
// Global event loop.

static void EventLoop() 
{
	EventType event;
	UInt16 error;
	do {
		EvtGetEvent(&event, evtWaitForever);
		
		if (!SysHandleEvent(&event)) {
			if (!MenuHandleEvent(0, &event, &error)) {
				if (!ApplicationEventHandler(&event)) {
					FrmDispatchEvent(&event);
				}
			}
		}
	}
	while (event.eType != appStopEvent);
}


//// PilotMain /////////////////////////////////////////////////////////
// It all starts here...

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) 
{
	Err err = 0;
	UInt32 feature;

	switch( launchCode ) 
	{
		case sysAppLaunchCmdSystemReset : 
		case sysAppLaunchCmdNormalLaunch :
			err = FtrGet(CREATOR,0,&feature);
			if( err == ftrErrNoSuchFeature )
			{
				UInt32 pref;
				UInt16 size;

				// Get preference for default mode
				pref = DEFAULTMODE;
				size = sizeof(pref);
				PrefGetAppPreferences(CREATOR,FTR_MODE,&pref,&size,true);
				FtrSet(CREATOR,FTR_MODE,pref);
				feature = pref;

				// Get preference for toggle on/off
				// make sure to come up with TS on before first normal launch
				pref = launchCode == sysAppLaunchCmdSystemReset;
				size = sizeof(pref);
				PrefGetAppPreferences(CREATOR,FTR_SIDETOGG,&pref,&size,false);
				FtrSet(CREATOR,FTR_SIDETOGG,pref);

				FtrSet(CREATOR,FTR_ROCKERTIME,0);
				FtrSet(CREATOR,FTR_SIDEHELD,0);
				FtrSet(CREATOR,FTR_PRESSTIME,0);
				FtrSet(CREATOR,FTR_STATE,0);
			}

			touchScreenMode(feature);

			if( launchCode != sysAppLaunchCmdNormalLaunch ) 
				return 0;

		
			CurrentMode = feature;

			err = StartApplication();
			if( err == 0 )
			{
				EventLoop();
				StopApplication();
			}

			break;

		case sysAppLaunchCmdNotify : 
			HandleNotification((SysNotifyParamType *)cmdPBP);
			break;

		default:
			break;
	}


	return err;
}

